'use strict';

let lastCondition;

/**
 * Appends content to an element
 *
 * @param {HTMLElement|SVGElement} element
 * @param {Object} content
 */
function append(element, content) {
    if(content instanceof Array) {
        for(let i in content) {
            append(element, content[i]);
        }

    } else if(content !== null && content !== undefined) {
        // Crisp view
        if(content instanceof Crisp.View) {
            append(element, content.$element || content.element);

        // jQuery logic
        } else if(typeof jQuery !== 'undefined') {
            if(element instanceof jQuery == false) {
                element = jQuery(element);
            }

            element.append(content);

        // Native JavaScript logic
        } else {
            if(typeof content === 'number' || typeof content === 'string') {
                element.innerHTML += content.toString();
            
            } else {
                element.appendChild(content);
        
            }
        }
    }
}

/**
 * Assigns event handler shorthands to element
 * This is done to prevent extending the HTMLElement/SVGElement prototype
 *
 * @param {HTMLElement|SVGElement} element
 */
function assignEvents(element) {
    /**
     * Handles the 'addEventListener' method
     *
     * @param {String} type
     * @param {Function} callback
     */
    element.on = function on(type, callback) {
        element.addEventListener(type, callback);

        return element;
    };

    /**
     * Handles the 'removeEventListener' method
     *
     * @param {String} type
     * @param {Function} callback
     */
    element.off = function off(type, callback) {
        element.removeEventListener(type, callback);

        return element;
    };

    /**
     * Removes an element
     */
    element.remove = function remove() {
        element.parentNode.removeChild(element);
    };

    // Define shorthand methods
    let shorthands = [
        'blur',
        'change',
        'click',
        'focus',
        'hover',
        'keydown',
        'keypress',
        'keyup',
        'mousedown',
        'mouseenter',
        'mouseleave',
        'mouseout',
        'mouseover',
        'mouseup',
        'select'
    ];

    for(let shorthand of shorthands) {
        element[shorthand] = (callback) => {
            return element.on(shorthand, callback);
        };
    } 
}

/**
 * Creates an element
 *
 * @param {String} tag
 * @param {String} namespace
 * @param {Object} attributes
 * @param {Object} contents
 *
 * @returns {HTMLElement|SVGElement} element
 */
function create(tag, namespace, attributes, contents) {
    let element = !!namespace ? document.createElementNS(namespace, tag) : document.createElement(tag);

    let isContents = (obj) => {
        if(obj === null || obj === undefined) { return false; }

        return obj instanceof Array ||
            obj instanceof HTMLElement || 
            obj instanceof SVGElement || 
            (typeof jQuery !== 'undefined' && obj instanceof jQuery) ||
            typeof obj === 'string' ||
            typeof obj === 'number';
    };

    // The attribute parameter is the content
    if(isContents(attributes)) {
        contents = [ attributes, contents ];   
    
    // The attribute parameter was defined as an object
    } else if (
        attributes !== undefined &&
        attributes instanceof Object &&
        !Array.isArray(attributes)
    ) {
        try {
            for(let key in attributes) {
                let value = attributes[key];

                // Null, undefined or false values should not be included
                if(!value && value !== 0) { continue; }

                // Shorthand attributes
                if(key === 'dynamicContent') {
                    key = 'data-cr-dynamic-content';
                } else if(key === 'dynamicAttributes') {
                    key = 'data-cr-dynamic-attributes';
                } else if(key === 'dynamicChildren') {
                    key = 'data-cr-dynamic-children';
                }

                element.setAttribute(key, value);
            }
        } catch(e) {
            console.log(e);
        }
    }
       
    append(element, contents);
    
    // jQuery logic
    if(typeof jQuery !== 'undefined') {
        return jQuery(element);

    // Native JavaScript logic
    } else {
        // Assign custom event functions to element instead of extending the prototype
        assignEvents(element);
        
        return element;
    }
}

class Elements {
    /**
     * Defines a new component
     *
     * @param {String} tag
     * @param {String} namespace
     * @param {Function} template
     */
    static component(tag, namespace, template) {
        Elements[tag] = (model) => {
            return create(tag, namespace, template(model));
        };
    }

    /**
     * Appends content using the function templating rules
     *
     * @params {HTMLElement|SVGElement} parentElement
     * @params {HTMLElement|SVGElement} contents
     */
    static append(parentElement, ...contents) {
        append(parentElement, contents);
    }
    
    /**
     * Replaces content using the function templating rules
     *
     * @params {HTMLElement|SVGElement} parentElement
     * @params {HTMLElement|SVGElement} contents
     */
    static replace(parentElement, ...contents) {
        parentElement.innerHTML = '';
        
        append(parentElement, contents);
    }

    /**
     * Encapsulates logic and renders the result
     *
     * @param {Function} func
     *
     * @param {HTMLElement|SVGElement} Contents
     */
    static do(func) {
        return func();
    }

    /**
     * Renders content based on a condition
     * 
     * @param {Boolean} condition
     * @param {HTMLElement|SVGElement} contents
     *
     * @returns {HTMLElement|SVGElement} Contents
     */
    static if(condition, ...contents) {
        lastCondition = condition || false;

        if(lastCondition) {
            lastCondition = undefined;

            return contents;
        }
    }

    /**
     * Uses the last provided condition to simulate an "else" statement
     *
     * @param {HTMLElement|SVGElement} contents
     *
     * @returns {HTMLElement|SVGElement} Contents
     */
    static else(...contents) {
        if(typeof lastCondition === 'undefined') {
            throw new Error('No "if" statement was provided before this "else" statement');
        }

        if(!lastCondition) {
            lastCondition = undefined;

            return contents;
        }
    }

    /**
     * Loops through an array or object, rendering elements from model data
     *
     * @param {Object} array
     * @param {Function} callback
     *
     * @returns {HTMLElement|SVGElement} elements
     */
    static each(array, callback) {
        var elements = [];

        for(var i in array) {
            var element = callback(i, array[i]);

            if(element) {
                elements.push(element);
            }
        }
         
        return elements;
    }

    /**
     * Loops a given number of times, rendering elements for each pass
     *
     * @param {Number} iterations
     * @param {Function} callback
     *
     * @returns {HTMLElement|SVGElement} Elements
     */
    static loop(iterations, callback) {
        var elements = [];

        for(var i = 0; i <= iterations; i++) {
            var element = callback(i);

            if(element) {
                elements.push(element);
            }
        }
         
        return elements;
    }

    /**
     * A shorthand for document.querySelector
     *
     * @param {String} query
     *
     * @returns {HTMLElement|SVGElement} Element
     */
    static find(query) {
        let element = document.querySelector(query);
        
        if(!element) { return null; }

        if(typeof jQuery !== 'undefined') {
            return jQuery(element);
        
        } else {
            assignEvents(element);

            return element;
        
        }
    }

    /**
     * A shorthand for document.querySelectorAll
     *
     * @param {String} query
     *
     * @returns {Array} Elements
     */
    static findAll(query) {
        let elements = document.querySelectorAll(query);
        
        if(!elements) { return []; }

        if(typeof jQuery !== 'undefined') {
            return jQuery(elements);
        
        } else {
            let array = [];

            for(let i = 0; i < elements.length; i++) {
                array[i] = elements[i];
                assignEvents(array[i]);
            }

            return array;
        }
    }
}

// ----------
// Init all element types
// ----------
function declareMethod(tag, namespace) {
    Elements[tag] = (attributes, ...contents) => {
        return create(tag, namespace, attributes, contents);
    };
}

let svgElementTypes = [
    'altGlyph',
    'altGlyphDef',
    'altGlyphItem',
    'animate',
    'animateColor',
    'animateMotion',
    'animateTransform',
    'circle',
    'clipPath',
    'color-profile',
    'cursor',
    'defs',
    'desc',
    'discard',
    'ellipse',
    'feBlend',
    'feColorMatrix',
    'feComponentTransfer',
    'feComposite',
    'feConvolveMatrix',
    'feDiffuseLighting',
    'feDisplacementMap',
    'feDistantLight',
    'feDropShadow',
    'feFlood',
    'feFuncA',
    'feFuncB',
    'feFuncG',
    'feFuncR',
    'feGaussianBlur',
    'feImage',
    'feMerge',
    'feMergeNode',
    'feMorphology',
    'feOffset',
    'fePointLight',
    'feSpecularLighting',
    'feSpotLight',
    'feTile',
    'feTurbulence',
    'filter',
    'font',
    'font-face',
    'font-face-format',
    'font-face-name',
    'font-face-src',
    'font-face-uri',
    'foreignObject',
    'g',
    'glyph',
    'glyphRef',
    'hatch',
    'hatchpath',
    'hkern',
    'image',
    'line',
    'linearGradient',
    'marker',
    'mask',
    'mesh',
    'meshgradient',
    'meshpatch',
    'meshrow',
    'metadata',
    'missing-glyph',
    'mpath',
    'path',
    'pattern',
    'polygon',
    'polyline',
    'radialGradient',
    'rect',
    'script',
    'set',
    'solidcolor',
    'stop',
    'svg',
    'switch',
    'symbol',
    'text',
    'textPath',
    'tref',
    'tspan',
    'unknown',
    'use',
    'view',
    'vkern'
];

for(var i in svgElementTypes) {
    declareMethod(svgElementTypes[i], 'http://www.w3.org/2000/svg');
}

let htmlElementTypes = [
    'a',
    'abbr',
    'address',
    'area',
    'article',
    'aside',
    'audio',
    'b',
    'base',
    'bdi',
    'bdo',
    'blockquote',
    'body',
    'br',
    'button',
    'canvas',
    'caption',
    'cite',
    'code',
    'col',
    'colgroup',
    'datalist',
    'dd',
    'del',
    'details',
    'dfn',
    'dialog',
    'div',
    'dl',
    'dt',
    'em',
    'embed',
    'fieldset',
    'figcaption',
    'figure',
    'footer',
    'form',
    'h1',
    'h2',
    'h3',
    'h4',
    'h5',
    'h6',
    'head',
    'header',
    'hr',
    'html',
    'i',
    'iframe',
    'img',
    'input',
    'ins',
    'kbd',
    'keygen',
    'label',
    'legend',
    'li',
    'link',
    'main',
    'map',
    'mark',
    'menu',
    'menuitem',
    'meta',
    'meter',
    'nav',
    'noscript',
    'object',
    'ol',
    'optgroup',
    'option',
    'output',
    'p',
    'param',
    'picture',
    'polygon',
    'pre',
    'progress',
    'q',
    'rp',
    'rt',
    'ruby',
    's',
    'samp',
    'script',
    'section',
    'select',
    'small',
    'source',
    'span',
    'strong',
    'style',
    'sub',
    'summary',
    'sup',
    'table',
    'tbody',
    'td',
    'textarea',
    'tfoot',
    'th',
    'thead',
    'time',
    'title',
    'tr',
    'track',
    'u',
    'ul',
    'var',
    'video',
    'wbr'
];

for(var i in htmlElementTypes) {
    declareMethod(htmlElementTypes[i]);
}

// Add to Crisp namespace
if(typeof Crisp === 'undefined') { window.Crisp = {}; }
Crisp.Elements = Elements;
