'use strict';

/**
 * Generates a new GUID
 */
function guid() {
    function s4() {
        return Math.floor((1 + Math.random()) * 0x10000)
            .toString(16)
            .substring(1);
    }

    return s4() + s4() + '-' + s4() + '-' + s4() + '-' +
        s4() + '-' + s4() + s4() + s4();
}

// Instances
const instances = {};

// Mutation observer
new MutationObserver((mutations) => {
    if(!mutations) { return; }

    for(let mutation of mutations) {
        // If any elements were removed, clean up views
        if(mutation.removedNodes.length > 0) {
            // Wait a few cycles before removing, as the element might just have been relocated
            setTimeout(() => {
                View.cleanup();
            });
            return;
        }
    }
}).observe(document.body, { childList: true, subtree: true });

/**
 * View class
 *
 * @class View
 *
 * @param {Object} params
 */
class View {
    constructor(params) {
        this.name = this.constructor.name;
        this.guid = guid();
        this.events = {};

        this._adopt(params);

        instances[this.guid] = this;

        // Placeholder element
        this.element = document.createElement('div');
        this.element.className = 'cr-placeholder';
        
        if(typeof jQuery !== 'undefined') {
            this.$element = $(this.element);
        }
    }

    /**
     * Backwards compatible extending
     */
    static extend(func) {
        func.prototype = Object.create(View.prototype);
        func.prototype.constructor = View;

        return func;
    }

    /**
     * Gets the name of this View
     */
    getName() {
        let name = this.constructor.toString();
        name = name.substring('function '.length);
        name = name.substring(0, name.indexOf('('));

        return name;
    }

    /**
     * Check if view should be removed
     */
    cleanup() {
        // Check if the element exists in the DOM
        let parentElement = this.element ? this.element.parentElement : null;

        while(parentElement != null) {
            if(parentElement === document.body) { return; }
            
            parentElement = parentElement.parentElement;
        }

        // Check if the element is the placeholder
        if(this.element && this.element.className === 'cr-placeholder') { return; }
        
        // If none of the above conditions were met, remove this instance
        this.remove();
    }

    /**
     * Init
     */
    _init() {
        this.prerender();
        this._render();
        this.postrender();
        
        if(!this.element && this.$element && this.$element.length > 0) {
            this.element = this.$element[0];   
        }
       
        if(!this.element) { return; }

        this.trigger('ready', this);
        this.isReady = true;
    }

    /**
     * Shorthand for .on('ready')
     */
    ready(callback) {
        if(this.isReady) {
            callback(this);
        } else {
            this.on('ready', callback);
        }
    }

    /**
     * Adopts values
     *
     * @param {Object} values
     */
    _adopt(params) {
        for(let k in params) {
            this[k] = params[k];
        }

        return this;
    }

    /**
     * Runs before render
     */
    prerender() {

    }

    /**
     * Tries to update dynamic elements only, but will fall back to a full render if needed
     */
    update() {
        if(!this.element) { return; }

        let output = this._getRenderOutput();
    
        let oldElements = this.element.querySelectorAll('*');
        let newElements = output.querySelectorAll('*');

        if(!oldElements || !newElements) { return; }

        // A flag to check if we should do a normal render
        let shouldRender = false;

        // Checks for dynamic children
        let checkChildren = (oldElement, newElement) => {
            if(!oldElement || !newElement) { return; }

            // The old element is not dynamic, skip it
            if(!oldElement.dataset.crDynamicChildren) { return; }

            // The old element is dynamic, but the new one isn't, do a normal render
            if(!newElement.dataset.crDynamicChildren) { return shouldRender = true; }

            // Both old and new elements at the same index are dynamic,
            // and the children have changed, so replace the old element with the new
            let oldChildren = oldElement.querySelectorAll('*');
            let newChildren = newElement.querySelectorAll('*');

            if(oldChildren.length !== newChildren.length) {
                oldElement.parentNode.replaceChild(newElement, oldElement);
            }
        }
        
        // Checks for dynamic content
        let checkContent = (oldElement, newElement) => {
            if(!oldElement || !newElement) { return; }

            // The old element is not dynamic, skip it
            if(!oldElement.dataset.crDynamicContent) { return; }

            // The old element is dynamic, but the new one isn't, do a normal render
            if(!newElement.dataset.crDynamicContent) { return shouldRender = true; }

            // Replace old content with new
            oldElement.parentNode.replaceChild(newElement, oldElement);
        }
        
        // Checks for dynamic attributes
        let checkAttributes = (oldElement, newElement) => {
            if(!oldElement || !newElement) { return; }

            // The old element is not dynamic, skip it
            if(!oldElement.dataset.crDynamicAttributes) { return; }

            // The old element is dynamic, but the new one isn't, do a normal render
            if(!newElement.dataset.crDynamicAttributes) { return shouldRender = true; }

            // Remove old attributes
            for(let i = 0; i < oldElement.attributes.length; i++) {
                oldElement.removeAttribute(oldElement.attributes[i].name, oldElement.attributes[i].value);
            }
            
            // Insert new attributes
            for(let i = 0; i < newElement.attributes.length; i++) {
                oldElement.setAttribute(newElement.attributes[i].name, newElement.attributes[i].value);
            }
        }

        // Check root elements
        checkContent(this.element, output);
        checkChildren(this.element, output);
        checkAttributes(this.element, output);
            
        if(shouldRender) { return this._render(); }

        // Check all child elements
        for(let i = 0; i < oldElements.length; i++) {
            checkContent(oldElements[i], newElements[i]);
            checkChildren(this.element, output);
            checkAttributes(oldElements[i], newElements[i]);

            if(shouldRender) { return this._render(); }
        }
    }

    /**
     * Gets the render output
     *
     * @returns {HTMLElement} Output
     */
    _getRenderOutput() {
        if(typeof this.template !== 'function') { return; }
            
        let output = this.template.call(this, this);

        if(!output) { return; }

        // In case of jQuery, convert to native HTMLElement
        if(typeof jQuery !== 'undefined' && output instanceof jQuery) {
            output = output[0];
        }

        return output;
    }

    /**
     * Renders this view
     */
    _render() {
        let output = this._getRenderOutput();

        if(!output) { return; }

        if(this.element.parentNode) {
            this.element.parentNode.replaceChild(output, this.element);
        }

        // jQuery
        if(typeof jQuery !== 'undefined') {
            this.$element = $(output);
            this.element = this.$element[0];

        // Native JavaScript
        } else {
            this.element = output;

        }

        this.element.dataset.guid = this.guid;
    }

    /**
     * Runs after render
     */
    postrender() {

    }

    /**
     * Removes the view from DOM and memory
     */
    remove(timeout) {
        if(!this.destroyed) {
            this.destroyed = true;

            setTimeout(() => {
                this.trigger('remove');

                // jQuery
                if(typeof jQuery !== 'undefined' && this.$element && this.$element.length > 0) {
                    this.$element.remove();

                // Native JavaScript
                } else if(this.element && this.element.parentElement) {
                    this.element.parentElement.removeChild(this.element);

                }

                delete instances[this.guid];
            }, timeout || 0 );
        }
    }

    /**
     * Call an event
     */
    _call(callback, data, ui) {
        if(typeof callback !== 'function') { return; }

        callback(data, ui, this);
    }

    /**
     * Trigger an event
     */
    trigger(e, obj) {
        if(this.events[e]) {
            if(typeof this.events[e] === 'function') {
                this.events[e](obj);
            } else {
                for(let i in this.events[e]) {
                    if(typeof this.events[e][i] === 'function') {
                        this.events[e][i](obj);
                    }
                }
            }
        }
    }

    /**
     * Bind an event
     */
    on(e, callback) {
        let view = this;

        // No events registered, register this as the only event
        if(!this.events[e]) {
            this.events[e] = function(data) { view._call(callback, data, this); };
        
        // Events have already been registered, add to callback array
        } else {
            // Only one event is registered, so convert from a single reference to an array
            if(!this.events[e].length) {
                this.events[e] = [
                    this.events[e]
                ];
            }
           
            // Insert the event call into the array 
            this.events[e].push(function(data) { view._call(callback, data, this); });
        }
    }

    /**
     * Check if event exists
     */
    hasEvent(name) {
        for(let k in this.events) {
            if(k == name) {
                return true;
            }
        }

        return false;
    }

    /**
     * Fetch model data
     */
    fetch() {
        let getModel = () => {
            // Get model from URL
            if(!this.model && typeof this.modelUrl === 'string') {
                let request = new XMLHttpRequest();
                request.open('GET', this.modelUrl, true);

                request.onload = () => {
                    if (request.status >= 200 && request.status < 400) {
                        // Success!
                        let data = JSON.parse(request.responseText);
                        
                        this.model = data;
                        
                        this._init();
                    } else {
                        // We reached our target server, but it returned an error
                        throw new Error('Couldn\'t fetch model data');
                    }
                };

                request.onerror = (e) => {
                    throw e;
                };

                request.send();
            
            // Get model with function
            } else if(!this.model && typeof this.modelFunction === 'function') {
                this.modelFunction(function(data) {
                    this.model = data;

                    this._init();
                });

            // Just perform the initialisation
            } else {
                this._init();
            }
        }
        
        // Get the model
        getModel();    
    }
   
    /**
     * Gets all instances of a view type
     *
     * @param {String|View} type
     */
    static getAll(type) {
        let results = [];

        if(type) {
            for(let i in instances) {
                let instance = instances[i];
                let isTypeMatch = typeof type === 'function' && instance instanceof type;
                let isStringMatch = typeof type === 'string' && instance.name === type;

                if(isTypeMatch || isStringMatch) {
                    results.push(instance);
                }
            }
        } else {
            results = Object.values(instances);
        }

        return results;
    }

    /**
     * Gets the first instance of a view type
     *
     * @param {String|View} type
     */
    static get(type) {
        let results = View.getAll(type);
        let view = results.length > 0 ? results[0] : null;

        return view;
    }

    /**
     * Removes all views of this type
     *
     * @param {String} type
     */
    static clear(type) {
        for(let guid in instances) {
            let instance = instances[guid];
            let isTypeMatch = typeof type === 'function' && instance instanceof type;
            let isStringMatch = typeof type === 'string' && instance.name === type;

            if(isTypeMatch || isStringMatch) {
                instance.remove();
            }
        }
    }
    
    /**
     * Cleans up orphaned views
     */
    static cleanup() {
        for(let guid in instances) {
            instances[guid].cleanup();
        }
    }
}

// Add to Crisp namespace
if(typeof Crisp === 'undefined') { window.Crisp = {}; }
Crisp.View = View;
